#pragma once

#include <gdk/gdk.h>

void do_compare     (int *argc, const char ***argv);
void do_convert     (int *argc, const char ***argv);
void do_info        (int *argc, const char ***argv);
void do_relabel     (int *argc, const char ***argv);
void do_show        (int *argc, const char ***argv);
void do_play        (int *argc, const char ***argv);

