/* This file is auto-generated; changes will not be preserved */
#pragma once

#include "gskgpuprintprivate.h"
#include "gskgpushaderopprivate.h"
#include "gskgradientprivate.h" /* for GskRepeat */
#include "gsk/gskroundedrectprivate.h"

void
gsk_gpu_convert_builtin_op (GskGpuRenderPass       *pass,
                            GdkColorState          *ccs,
                            GdkColorState          *acs,
                            const graphene_rect_t  *bounds,
                            GskGpuImage            *image1,
                            GskGpuSampler           sampler1,
                            GdkBuiltinColorStateId  variation_color_space,
                            gboolean                variation_premultiply,
                            gboolean                variation_reverse,
                            const graphene_rect_t  *tex_rect);
