#include</build/gtk4/src/build/gtk/svg/gtksvgenumtypes.c>
#include</build/gtk4/src/gtk/gtk/svg/gtksvg.c>
#include</build/gtk4/src/gtk/gtk/svg/gtksvgclip.c>
#include</build/gtk4/src/gtk/gtk/svg/gtksvgcolor.c>
#include</build/gtk4/src/gtk/gtk/svg/gtksvgcolorstop.c>
#include</build/gtk4/src/gtk/gtk/svg/gtksvgcontentfit.c>
#include</build/gtk4/src/gtk/gtk/svg/gtksvgdasharray.c>
#include</build/gtk4/src/gtk/gtk/svg/gtksvgelement.c>
#include</build/gtk4/src/gtk/gtk/svg/gtksvgelementtype.c>
#include</build/gtk4/src/gtk/gtk/svg/gtksvgenum.c>
#include</build/gtk4/src/gtk/gtk/svg/gtksvgerror.c>
#include</build/gtk4/src/gtk/gtk/svg/gtksvgeventcontroller.c>
#include</build/gtk4/src/gtk/gtk/svg/gtksvgfilter.c>
#include</build/gtk4/src/gtk/gtk/svg/gtksvgfilterfunctions.c>
#include</build/gtk4/src/gtk/gtk/svg/gtksvgfilterref.c>
#include</build/gtk4/src/gtk/gtk/svg/gtksvgfiltertype.c>
#include</build/gtk4/src/gtk/gtk/svg/gtksvggpa.c>
#include</build/gtk4/src/gtk/gtk/svg/gtksvghref.c>
#include</build/gtk4/src/gtk/gtk/svg/gtksvgkeyword.c>
#include</build/gtk4/src/gtk/gtk/svg/gtksvglanguage.c>
#include</build/gtk4/src/gtk/gtk/svg/gtksvglocation.c>
#include</build/gtk4/src/gtk/gtk/svg/gtksvgmask.c>
#include</build/gtk4/src/gtk/gtk/svg/gtksvgmediaquery.c>
#include</build/gtk4/src/gtk/gtk/svg/gtksvgnumber.c>
#include</build/gtk4/src/gtk/gtk/svg/gtksvgnumbers.c>
#include</build/gtk4/src/gtk/gtk/svg/gtksvgorient.c>
#include</build/gtk4/src/gtk/gtk/svg/gtksvgpaint.c>
#include</build/gtk4/src/gtk/gtk/svg/gtksvgpangoutils.c>
#include</build/gtk4/src/gtk/gtk/svg/gtksvgparser.c>
#include</build/gtk4/src/gtk/gtk/svg/gtksvgpath.c>
#include</build/gtk4/src/gtk/gtk/svg/gtksvgpathdata.c>
#include</build/gtk4/src/gtk/gtk/svg/gtksvgpathutils.c>
#include</build/gtk4/src/gtk/gtk/svg/gtksvgproperty.c>
#include</build/gtk4/src/gtk/gtk/svg/gtksvgrenderer.c>
#include</build/gtk4/src/gtk/gtk/svg/gtksvgserialize.c>
#include</build/gtk4/src/gtk/gtk/svg/gtksvgstring.c>
#include</build/gtk4/src/gtk/gtk/svg/gtksvgstringlist.c>
#include</build/gtk4/src/gtk/gtk/svg/gtksvgstringutils.c>
#include</build/gtk4/src/gtk/gtk/svg/gtksvgtextdecoration.c>
#include</build/gtk4/src/gtk/gtk/svg/gtksvgtimespec.c>
#include</build/gtk4/src/gtk/gtk/svg/gtksvgtransform.c>
#include</build/gtk4/src/gtk/gtk/svg/gtksvgunit.c>
#include</build/gtk4/src/gtk/gtk/svg/gtksvgutils.c>
#include</build/gtk4/src/gtk/gtk/svg/gtksvgvalue.c>
#include</build/gtk4/src/gtk/gtk/svg/gtksvgviewbox.c>
#include</build/gtk4/src/gtk/gtk/svg/gtksvganimation.c>
