/* This file is auto-generated; changes will not be preserved */
#pragma once

#include "gskgpuprintprivate.h"
#include "gskgpushaderopprivate.h"
#include "gskgradientprivate.h" /* for GskRepeat */
#include "gsk/gskroundedrectprivate.h"

void
gsk_gpu_border_op (GskGpuRenderPass      *pass,
                   GdkColorState         *ccs,
                   GdkColorState         *acs,
                   const graphene_rect_t *bounds,
                   const GskRoundedRect  *outline,
                   const GdkColor        *top_border_color,
                   const GdkColor        *right_border_color,
                   const GdkColor        *bottom_border_color,
                   const GdkColor        *left_border_color,
                   const graphene_vec4_t *border_widths,
                   const graphene_size_t *inside_offset);
