
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#include "config.h"
#include "gtksvgenumtypes.h"
#include <gtksvg.h>
/* enumerations from "gtksvg.h" */
GType
gtk_svg_features_get_type (void)
{
  static size_t g_define_type_id__volatile = 0;

  if (g_once_init_enter (&g_define_type_id__volatile))
    {
      static const GFlagsValue values[] = {
        { GTK_SVG_ANIMATIONS, "GTK_SVG_ANIMATIONS", "animations" },
        { GTK_SVG_SYSTEM_RESOURCES, "GTK_SVG_SYSTEM_RESOURCES", "system-resources" },
        { GTK_SVG_EXTERNAL_RESOURCES, "GTK_SVG_EXTERNAL_RESOURCES", "external-resources" },
        { GTK_SVG_EXTENSIONS, "GTK_SVG_EXTENSIONS", "extensions" },
        { GTK_SVG_TRADITIONAL_SYMBOLIC, "GTK_SVG_TRADITIONAL_SYMBOLIC", "traditional-symbolic" },
        { 0, NULL, NULL }
      };
      GType g_define_type_id =
        g_flags_register_static (g_intern_static_string ("GtkSvgFeatures"), values);
      g_once_init_leave (&g_define_type_id__volatile, g_define_type_id);
    }

  return g_define_type_id__volatile;
}
GType
gtk_svg_error_get_type (void)
{
  static size_t g_define_type_id__volatile = 0;

  if (g_once_init_enter (&g_define_type_id__volatile))
    {
      static const GEnumValue values[] = {
        { GTK_SVG_ERROR_INVALID_SYNTAX, "GTK_SVG_ERROR_INVALID_SYNTAX", "invalid-syntax" },
        { GTK_SVG_ERROR_INVALID_ELEMENT, "GTK_SVG_ERROR_INVALID_ELEMENT", "invalid-element" },
        { GTK_SVG_ERROR_INVALID_ATTRIBUTE, "GTK_SVG_ERROR_INVALID_ATTRIBUTE", "invalid-attribute" },
        { GTK_SVG_ERROR_MISSING_ATTRIBUTE, "GTK_SVG_ERROR_MISSING_ATTRIBUTE", "missing-attribute" },
        { GTK_SVG_ERROR_INVALID_REFERENCE, "GTK_SVG_ERROR_INVALID_REFERENCE", "invalid-reference" },
        { GTK_SVG_ERROR_FAILED_UPDATE, "GTK_SVG_ERROR_FAILED_UPDATE", "failed-update" },
        { GTK_SVG_ERROR_FAILED_RENDERING, "GTK_SVG_ERROR_FAILED_RENDERING", "failed-rendering" },
        { GTK_SVG_ERROR_IGNORED_ELEMENT, "GTK_SVG_ERROR_IGNORED_ELEMENT", "ignored-element" },
        { GTK_SVG_ERROR_LIMITS_EXCEEDED, "GTK_SVG_ERROR_LIMITS_EXCEEDED", "limits-exceeded" },
        { GTK_SVG_ERROR_NOT_IMPLEMENTED, "GTK_SVG_ERROR_NOT_IMPLEMENTED", "not-implemented" },
        { GTK_SVG_ERROR_FEATURE_DISABLED, "GTK_SVG_ERROR_FEATURE_DISABLED", "feature-disabled" },
        { 0, NULL, NULL }
      };
      GType g_define_type_id =
        g_enum_register_static (g_intern_static_string ("GtkSvgError"), values);
      g_once_init_leave (&g_define_type_id__volatile, g_define_type_id);
    }

  return g_define_type_id__volatile;
}

/* Generated data ends here */

