/* This file is auto-generated; changes will not be preserved */
#pragma once

#include "gskgpuprintprivate.h"
#include "gskgpushaderopprivate.h"
#include "gskgradientprivate.h" /* for GskRepeat */
#include "gsk/gskroundedrectprivate.h"

void
gsk_gpu_arithmetic_op (GskGpuRenderPass      *pass,
                       GdkColorState         *ccs,
                       GdkColorState         *acs,
                       const graphene_rect_t *bounds,
                       GskGpuImage           *image1,
                       GskGpuSampler          sampler1,
                       GskGpuImage           *image2,
                       GskGpuSampler          sampler2,
                       const graphene_rect_t *first_rect,
                       const graphene_rect_t *second_rect,
                       float                  k1,
                       float                  k2,
                       float                  k3,
                       float                  k4);
