/* This file is auto-generated; changes will not be preserved */
#pragma once

#include "gskgpuprintprivate.h"
#include "gskgpushaderopprivate.h"
#include "gskgradientprivate.h" /* for GskRepeat */
#include "gsk/gskroundedrectprivate.h"

void
gsk_gpu_radial_gradient_op (GskGpuRenderPass       *pass,
                            GdkColorState          *ccs,
                            GdkColorState          *acs,
                            const graphene_rect_t  *bounds,
                            gboolean                variation_supersampling,
                            gboolean                variation_concentric,
                            gboolean                variation_premultiplied,
                            GskRepeat               variation_repeat,
                            const GdkColor         *color0,
                            const GdkColor         *color1,
                            const GdkColor         *color2,
                            const GdkColor         *color3,
                            const GdkColor         *color4,
                            const GdkColor         *color5,
                            const GdkColor         *color6,
                            const graphene_vec4_t  *offsets0,
                            const graphene_vec4_t  *offsets1,
                            const graphene_vec4_t  *hints0,
                            const graphene_vec4_t  *hints1,
                            const graphene_point_t *start_center,
                            const graphene_size_t  *start_radius,
                            const graphene_point_t *end_center,
                            const graphene_size_t  *end_radius);
