/* This file is auto-generated; changes will not be preserved */
#pragma once

#include "gskgpuprintprivate.h"
#include "gskgpushaderopprivate.h"
#include "gskgradientprivate.h" /* for GskRepeat */
#include "gsk/gskroundedrectprivate.h"

void
gsk_gpu_color_op (GskGpuRenderPass      *pass,
                  GdkColorState         *ccs,
                  GdkColorState         *acs,
                  const graphene_rect_t *bounds,
                  const GdkColor        *color);
