/* This file is auto-generated; changes will not be preserved */
#pragma once

#include "gskgpuprintprivate.h"
#include "gskgpushaderopprivate.h"
#include "gskgradientprivate.h" /* for GskRepeat */
#include "gsk/gskroundedrectprivate.h"

void
gsk_gpu_blur_op (GskGpuRenderPass      *pass,
                 GdkColorState         *ccs,
                 GdkColorState         *acs,
                 const graphene_rect_t *bounds,
                 GskGpuImage           *image1,
                 GskGpuSampler          sampler1,
                 gboolean               variation_colorize,
                 const graphene_rect_t *rect,
                 const GdkColor        *blur_color,
                 const graphene_rect_t *tex_rect,
                 const graphene_size_t *blur_direction);
