/* This file is auto-generated; changes will not be preserved */
#pragma once

#include "gskgpuprintprivate.h"
#include "gskgpushaderopprivate.h"
#include "gskgradientprivate.h" /* for GskRepeat */
#include "gsk/gskroundedrectprivate.h"

void
gsk_gpu_box_shadow_op (GskGpuRenderPass      *pass,
                       GdkColorState         *ccs,
                       GdkColorState         *acs,
                       const graphene_rect_t *bounds,
                       gboolean               variation_inset,
                       const GskRoundedRect  *outline,
                       const graphene_size_t *shadow_offset,
                       float                  shadow_spread,
                       float                  blur_radius,
                       const GdkColor        *color);
